import pandas as pd
import pandas.testing as pdt
import pytest
from mongoengine import connect

from antarctic.pandasdocument import PandasDocument, NotUniqueError
from antarctic.timeutil import merge
from test.config import read_pd

client = connect(db="test", host="mongomock://localhost")


class Singer(PandasDocument):
    pass


class TestEngine(object):
    def test_document(self):
        # Create a new page and add tags
        p = Singer(name="Peter Maffay")
        assert p.name == "Peter Maffay"

    def test_lt(self):
        assert Singer(name="A") < Singer(name="B")

    def test_reference(self):
        p = Singer(name="Peter Maffay")
        assert p.reference.get("NoNoNo", default=5) == 5

        p.reference["XXX"] = 10
        assert p.reference.keys() == {"XXX"}
        assert {k: v for k, v in p.reference.items()} == {"XXX": 10}

    def test_equals(self):
        # can't harm to clean a bit
        Singer.objects.delete()

        p1 = Singer(name="Peter Maffay")
        p2 = Singer(name="Peter Maffay")

        assert p1 == p2

    def test_merge(self):
        # can't harm to clean a bit
        Singer.objects.delete()

        p = Singer(name="Peter Maffay")

        ts1 = pd.Series(index=[1, 2], data=[3.3, 4.3])
        ts2 = pd.Series(index=[2, 3], data=[5.3, 6.3])

        p.close = ts1
        p.close = merge(old=p.close, new=ts2)

        pdt.assert_series_equal(p.close, pd.Series(index=[1, 2, 3], data=[3.3, 5.3, 6.3]))

    def test_products(self):
        # can't harm to clean a bit
        Singer.objects.delete()

        p1 = Singer(name="Peter").save()
        p2 = Singer(name="Falco").save()

        # here we query the database! Hence need the client in the background
        a = Singer.products(names=["Peter"])
        assert len(a) == 1
        assert a[0] == p1

        b = Singer.products()
        assert len(b) == 2
        assert set(b) == {p1, p2}

        frame = Singer.reference_frame(products=[p1, p2])
        assert set(frame.index) == {"Peter", "Falco"}
        assert frame.empty

        frame = Singer.reference_frame()
        assert set(frame.index) == {"Peter", "Falco"}
        assert frame.empty

    def test_reference_frame(self):
        # can't harm to clean a bit
        Singer.objects.delete()

        p1 = Singer(name="Peter")
        p1.reference["A"] = 20.0
        p1.save()

        p2 = Singer(name="Falco")
        p2.reference["A"] = 30.0
        p2.reference["B"] = 10.0
        p2.save()

        frame = Singer.reference_frame(products=[p1, p2])
        assert set(frame.index) == {"Peter", "Falco"}
        assert set(frame.keys()) == {"A", "B"}

        frame = Singer.reference_frame()
        assert set(frame.index) == {"Peter", "Falco"}
        assert set(frame.keys()) == {"A", "B"}

    def test_ts(self):
        # can't harm to clean a bit
        Singer.objects.delete()

        p1 = Singer(name="Peter Maffay")
        p1.price = pd.Series(data=[2, 3, 5], index=[pd.Timestamp("2010-01-01"), pd.Timestamp("2010-01-03"), pd.Timestamp("2010-01-05")])
        p1.save()

        p2 = Singer(name="Falco").save()
        assert p1.reference == {}

        frame = Singer.pandas_frame(item="price", products=[p1, p2])
        assert set(frame.keys()) == {"Peter Maffay"}
        pdt.assert_series_equal(p1.price, frame["Peter Maffay"], check_names=False)

        frame = Singer.pandas_frame(item="price")
        assert set(frame.keys()) == {"Peter Maffay"}
        pdt.assert_series_equal(p1.price, frame["Peter Maffay"], check_names=False)

    def test_pandas_wrong(self):
        # can't harm to clean a bit
        Singer.objects.delete()

        p1 = Singer(name="Peter Maffay")
        p1.price = 5.0
        assert Singer.pandas_frame(item="price", products=[p1]).empty

    def test_repr(self):
        p1 = Singer(name="Peter Maffay")
        assert str(p1) == "<Singer: Peter Maffay>"

    def test_not_unique_name(self):
        # can't harm to clean a bit
        Singer.objects.delete()

        # check that no singer has survived
        s = [singer for singer in Singer.objects]
        assert len(s) == 0

        c1 = Singer(name="AA").save()
        assert c1.name == "AA"

        # try to create a second singer with the same name
        with pytest.raises(NotUniqueError):
            Singer(name="AA").save()

    def test_to_dict(self):
        # can't harm to clean a bit
        Singer.objects.delete()

        c1 = Singer(name="AAA").save()
        c2 = Singer(name="BBB").save()

        assert Singer.to_dict() == {"AAA": c1, "BBB": c2}

    def test_frame(self):
        frame = read_pd("reference.csv", index_col=[0,1])
        s = Singer(name="Peter Maffay")
        s.frame = frame
        pdt.assert_frame_equal(frame, s.frame)

    def test_frame_timeseries(self):
        frame = read_pd("price.csv", index_col=0)
        s = Singer(name="Peter Maffay")
        s.frame = frame
        pdt.assert_frame_equal(frame, s.frame)


import datetime
import pandas as pd

from mongoengine import *

# We rely heavily on MongoEngine, see http://mongoengine.org/
# MongoEngine is an ORM for MongoDB, as sqlalchemy is an ORM for SQL
# MongoDB is a document based database.
# There is a one-to-one relationship between documents and objects.
# The objects are defined using MongoEngine


# An document could be
# {"Name": "IBM US Equity", "PX_LAST": 110.0, ...}

# Each document is dynamic and stores all time series and reference data related to an object
# Each object and therefore each document is a child of the abstract PandasDocument object.
from numpy import float64


class PandasDocument(DynamicDocument):
    # A PandasDocument is an abstract Mongo Document, e.g. instances of this document can not be instantiated.
    # All concrete objects such as Symbols or Strategies are children of the PandasDocument.
    # Having a common parent helps to share functionality
    meta = {'abstract': True}
    # Each children has a unique name (within the class of the children)
    name = StringField(max_length=200, required=True, unique=True)
    # A dictionary for reference data, e.g. something like symbol.reference["PX_LAST"] = ...
    reference = DictField()
    # Date modified
    date_modified = DateTimeField(default=datetime.datetime.utcnow)

    @classmethod
    def reference_frame(cls, products=None) -> pd.DataFrame:
        products = products or cls.objects

        frame = pd.DataFrame(
            {product.name: pd.Series({key: data for key, data in product.reference.items()}) for product in
             products}).transpose()
        frame.index.name = cls.__name__.lower()
        return frame.sort_index()

    def __lt__(self, other):
        # sort documents by name
        return self.name < other.name

    def __eq__(self, other):
        # two documents are the sname if they have the same name and class
        return self.__class__ == other.__class__ and self.name == other.name

    # we want to make a set of assets, etc....
    def __hash__(self):
        return hash(self.to_json())

    def __setattr__(self, key, value):
        if isinstance(value, pd.Series):
            DynamicDocument.__setattr__(self, key, value.to_json(orient="split"))
        elif isinstance(value, pd.DataFrame):
            DynamicDocument.__setattr__(self, key, value.to_json(orient="table"))
        else:
            DynamicDocument.__setattr__(self, key, value)

    def __getattribute__(self, item):
        if item.startswith("_"):
            return DynamicDocument.__getattribute__(self, item)

        x = DynamicDocument.__getattribute__(self, item)

        try:
            return pd.read_json(x, orient="table", typ="frame")
        except:
            pass

        try:
            return pd.read_json(x, orient="split", typ="series")
        except:
            pass

        return x

    @classmethod
    def products(cls, names=None):
        # extract symbols from database
        if names is None:
            return cls.objects
        else:
            return cls.objects(name__in=names)

    @classmethod
    def pandas_frame(cls, item, products=None) -> pd.DataFrame:
        products = products or cls.objects
        frame = pd.DataFrame({product.name: product.__pandas(item=item, default=pd.Series({}, dtype=float64)) for product in products})
        frame = frame.dropna(axis=1, how="all").transpose()
        frame.index.name = cls.__name__.lower()
        return frame.sort_index().transpose()

    def __pandas(self, item, default=None):
        try:
            obj = self.__getattribute__(item)
            if isinstance(obj, pd.Series) or isinstance(obj, pd.DataFrame):
                return obj

            raise AttributeError()
        except AttributeError:
            return default

    def __str__(self):
        return "<{type}: {name}>".format(type=self.__class__.__name__, name=self.name)

    @classmethod
    def to_dict(cls):
        # represent all documents of a class as a dictionary
        return {x.name: x for x in cls.objects}



import pandas as pd


def merge(new, old=None):
    # very smart merging here, new and old merge
    if new is not None:
        if old is not None:
            x = pd.concat((new, old), sort=True)
            return x.groupby(x.index).first().sort_index()
        else:
            return new

    else:
        return old


import pytest

from antarctic.timeutil import merge
import pandas.testing as pt
import pandas as pd
from test.config import read_pd


@pytest.fixture()
def ts():
    return read_pd("ts.csv", squeeze=True, header=None, parse_dates=True, index_col=0)


def test_merge(ts):
    x = merge(new=ts)
    pt.assert_series_equal(x, ts)

    x = merge(new=ts, old=ts)
    pt.assert_series_equal(x, ts)

    x = merge(new=5 * ts, old=ts)
    pt.assert_series_equal(x, 5 * ts)

    y = merge(None)
    assert not y

    y = merge(pd.Series({}), None)
    pt.assert_series_equal(y, pd.Series({}))

    y = merge(pd.Series({}), pd.Series({}))
    pt.assert_series_equal(y, pd.Series({}))